"""
Provides luigi Task subclasses that wrap the lib5c command line functions.
"""

import time
import os
import subprocess
import uuid

import luigi

from lib5c.util.system import shell_quote

try:
    from bsub import bsub

    bsub_avail = True
except ImportError:
    bsub_avail = False


def get_all_lines(filename):
    with open(filename, 'r') as handle:
        return handle.read()


def parallelize_reps(task_class, reps, **kwargs):
    return type(
        '%sWrapper_%s' % (task_class.__name__, str(uuid.uuid4())[:8]),
        (luigi.WrapperTask,),
        {'requires': lambda self: [task_class(rep=rep, **kwargs)
                                   for rep in reps]})


def parallelize_reps_regions(task_class, reps, regions, **kwargs):
    return type(
        '%sWrapper_%s' % (task_class.__name__, str(uuid.uuid4())[:8]),
        (luigi.WrapperTask,),
        {'requires': lambda self: [task_class(rep=rep, region=region, **kwargs)
                                   for rep in reps
                                   for region in regions]})


def add_visualization_hooks(f, pvalue=False, obs_over_exp=False, tetris=False):
    def hooked(self):
        f(self)
        if self.heatmap:
            if type(self.output()) in [list, tuple]:
                if self.heatmap_outdir:
                    outfile = os.path.join(self.heatmap_outdir, '%s_%r.png')
                else:
                    outfile = os.path.join(
                        os.path.split(self.output()[0].path)[0], '%s_%r.png')
                input_paths = [i.path for i in self.output()
                               if i.path.endswith('.counts')]
                if len(input_paths) == 1:
                    infile = input_paths[0]
                    outfile = infile.replace('.counts', '_%r.png')
                else:
                    common_prefix = os.path.commonprefix(input_paths)
                    common_postfix = os.path.commonprefix(
                        [p[::-1] for p in input_paths])[::-1]
                    infile = common_prefix + '*' + common_postfix
                cmd = 'lib5c plot heatmap -p %s -R %s %s' % \
                      (self.input()[0].path, shell_quote(infile), outfile)
            else:
                if self.heatmap_outdir:
                    outfile = os.path.join(
                        self.heatmap_outdir,
                        os.path.split(self.output().path)[0].replace('.counts',
                                                                     '_%r.png'))
                else:
                    outfile = self.output().path.replace('.counts', '_%r.png')
                cmd = 'lib5c plot heatmap -p %s -R %s %s' % \
                      (self.input()[0].path, self.output().path, outfile)
            if pvalue:
                cmd += ' -P'
            elif obs_over_exp:
                cmd += ' -c obs_over_exp'
            elif tetris:
                cmd += ' -TC'
            if bsub_avail:
                job_name = '.'.join([self.__class__.__name__] +
                                    ['%s_%s' % (k, v)
                                     for k, v in self.param_kwargs.iteritems()
                                     if not hasattr(v, '__getitem__')] +
                                    ['heatmap'])
                if len(job_name) > 250:
                    job_name = '%s.%s' % (self.__class__.__name__, 'heatmap')
                sub = bsub(job_name, verbose=False)
                sub(cmd)
            else:
                print(cmd)
                subprocess.call(cmd, shell=True)

    return hooked


def visualizable(pvalue=False, obs_over_exp=False, tetris=False):
    def visualizable_decorator(cls):
        setattr(cls, 'heatmap', luigi.BoolParameter(default=False))
        setattr(cls, 'heatmap_outdir', luigi.Parameter(default=None))
        setattr(cls, 'run', add_visualization_hooks(
            cls.run, pvalue=pvalue, obs_over_exp=obs_over_exp, tetris=tetris))
        return cls
    return visualizable_decorator


class CmdTask(luigi.Task):
    def _construct_cmd_string(self):
        raise NotImplementedError()

    def run(self):
        for output in luigi.task.flatten(self.output()):
            output.makedirs()
        if bsub_avail:
            if not os.path.exists('logs'):
                os.makedirs('logs')
            job_name = '.'.join([self.__class__.__name__] +
                                ['%s_%s' % (k, v)
                                 for k, v in self.param_kwargs.iteritems()
                                 if not hasattr(v, '__getitem__')])
            if len(job_name) > 250:
                job_name = self.__class__.__name__
            sub = bsub(job_name, verbose=True)
            sub(self._construct_cmd_string())
            bsub.poll(sub.job_id)
            time.sleep(10)
            for output in luigi.task.flatten(self.output()):
                if not output.exists():
                    raise Exception(
                        'missing file: %s\nerror:\n%s\ncommand:\n%s' %
                        (output.path,
                         get_all_lines(sub.kwargs['e'].replace('%J',
                                                               sub.job_id)),
                         self._construct_cmd_string()))
        else:
            cmd = self._construct_cmd_string()
            print(cmd)
            subprocess.call(cmd, shell=True)


class RegionalTaskMixin(object):
    """
    Mixin class for Tasks that write a separate output file per region.
    """
    region = luigi.Parameter(default=None)

    def _outfile_pattern(self):
        """
        This function should return a outfile pattern that may be parameterized
        by ``_parameterized_outfile_pattern()`` and then regionalized by
        ``_regional_outfiles()``.

        Returns
        -------
        str
            The outfile pattern.
        """
        raise NotImplementedError()

    def _parameterized_outfile_pattern(self):
        """
        Plugs in parameters from this instance into the outfile pattern.

        The resulting string will be passed to the tool in the command string.

        The default implementation is to not plug in any parameters.

        Returns
        -------
        str
            The parametrized outfile pattern.
        """
        return self._outfile_pattern()

    def _regional_outfiles(self, regions=None):
        """
        Creates a list of actual expected outfiles, which can be used to
        construct the output() definition, using the outfile pattern and the
        parameters of this instance.

        For example::

            def output(self):
                return [luigi.LocalTarget(outfile)
                        for outfile in self._regional_outfiles(regions=[...])]

        Params
        ------
        regions : Optional[List[str]]
            The regions to deduce the outfiles for.

        Returns
        -------
        List[str]
            The list of actual expected outfiles.
        """
        # baseline outfile from pattern
        outfile = self._parameterized_outfile_pattern()

        # honor self.region field if present
        if self.region is not None:
            regions = [self.region]

        # parallelize across regions
        if regions is not None:
            return [outfile.replace('%r', region) for region in regions]
        else:
            return [outfile]


@visualizable()
class OutliersTask(CmdTask):
    fold_threshold = luigi.FloatParameter(default=8.0)
    window_size = luigi.IntParameter(default=5)
    overwrite_value = luigi.Parameter(default='nan')

    def _construct_cmd_string(self):
        cmd = 'lib5c outliers -p %s -f %f -w %i -o %s' % \
              (self.input()[0].path, self.fold_threshold, self.window_size,
               self.overwrite_value)
        cmd += ' %s %s' % (self.input()[1].path, self.output().path)
        return cmd


@visualizable()
class QnormTask(CmdTask):
    outfile_pattern = luigi.Parameter(default='qnorm/%s_qnorm.counts')
    regional = luigi.BoolParameter(default=False)
    averaging = luigi.BoolParameter(default=False)
    condition_on = luigi.Parameter(default=None)
    reference = luigi.Parameter(default=None)

    def _construct_cmd_string(self):
        cmd = 'lib5c qnorm -p %s' % self.input()[0].path
        if self.regional:
            cmd += ' -R'
        if self.averaging:
            cmd += ' -A'
        if self.condition_on is not None:
            cmd += ' -c %s' % self.condition_on
        if self.reference is not None:
            cmd += ' -r %s' % self.reference
        cmd += ' %s %s' % (self.outfile_pattern,
                           ' '.join(i.path for i in self.input()[1:]))
        return cmd


@visualizable()
class ExpressTask(CmdTask):
    bias = luigi.BoolParameter(default=True)

    def _construct_cmd_string(self):
        cmd = 'lib5c express -p %s' % self.input()[0].path
        if self.bias:
            cmd += ' -B'
        cmd += ' %s %s' % (self.input()[1].path, self.output().path)
        return cmd


@visualizable()
class KnightRuizTask(CmdTask):
    bias = luigi.BoolParameter(default=True)
    imputation_size = luigi.IntParameter(default=0)

    def _construct_cmd_string(self):
        cmd = 'lib5c kr -p %s' % self.input()[0].path
        if self.bias:
            cmd += ' -B'
        if self.imputation_size is not None:
            cmd += ' -s %s' % self.imputation_size
        cmd += ' %s %s' % (self.input()[1].path, self.output().path)
        return cmd


@visualizable()
class IcedTask(CmdTask):
    bias = luigi.BoolParameter(default=True)
    imputation_size = luigi.IntParameter(default=0)

    def _construct_cmd_string(self):
        cmd = 'lib5c iced -p %s' % self.input()[0].path
        if self.bias:
            cmd += ' -B'
        if self.imputation_size is not None:
            cmd += ' -s %s' % self.imputation_size
        cmd += ' %s %s' % (self.input()[1].path, self.output().path)
        return cmd


@visualizable()
class SplineTask(CmdTask):
    bias_factors = luigi.ListParameter(default=['% GC', 'length'])
    knots = luigi.ListParameter(default=[0, 20])
    model_outfile = luigi.Parameter(default=None)

    def _construct_cmd_string(self):
        cmd = 'lib5c spline -p %s' % self.input()[0].path
        if self.bias_factors is not None:
            cmd += ' -b %s' % shell_quote('(%s)' % ','.join(self.bias_factors))
        if self.knots is not None:
            cmd += ' -k %s' % \
                   shell_quote('(%s)' % ','.join([str(k) for k in self.knots]))
        cmd += ' %s %s' % (self.input()[1].path, self.output().path)
        return cmd


class DetermineBinsTask(CmdTask):
    bin_width = luigi.IntParameter(default=4000)

    def _construct_cmd_string(self):
        return 'lib5c determine-bins -w %i %s %s' % \
               (self.bin_width, self.input().path, self.output().path)


class FilteringTask(CmdTask):
    window_width = luigi.IntParameter(default=20000)
    window_function = luigi.Parameter(default='gmean')
    threshold = luigi.FloatParameter(default=0.0)
    inverse_weights = luigi.BoolParameter(default=False)
    wipe_unsmoothable_columns = luigi.BoolParameter(default=True)

    def _construct_cmd_string(self):
        raise NotImplementedError()


@visualizable()
class BinTask(FilteringTask):
    def _construct_cmd_string(self):
        cmd = 'lib5c bin -b %s -p %s -f %s -w %i -t %f' % \
              (self.input()[0].path, self.input()[1].path, self.window_function,
               self.window_width, self.threshold)
        if self.inverse_weights:
            cmd += ' -I'
        if self.wipe_unsmoothable_columns:
            cmd += ' -W'
        cmd += ' %s %s' % (self.input()[2].path, self.output().path)
        return cmd


@visualizable()
class SmoothTask(FilteringTask):
    def _construct_cmd_string(self):
        cmd = 'lib5c smooth -p %s -f %s -w %i -t %f' % \
              (self.input()[0].path, self.window_function,
               self.window_width, self.threshold)
        if self.inverse_weights:
            cmd += ' -I'
        if self.wipe_unsmoothable_columns:
            cmd += ' -W'
        cmd += ' %s %s' % (self.input()[1].path, self.output().path)
        return cmd


@visualizable()
class ExpectedTask(CmdTask):
    monotonic = luigi.BoolParameter(default=False)
    exclude_near_diagonal = luigi.BoolParameter(default=True)
    powerlaw = luigi.BoolParameter(default=False)
    regression = luigi.BoolParameter(default=False)
    degree = luigi.IntParameter(default=1)
    lowess = luigi.BoolParameter(default=False)
    lowess_frac = luigi.FloatParameter(default=0.5)
    donut = luigi.BoolParameter(default=False)
    p = luigi.IntParameter(default=5)
    w = luigi.IntParameter(default=15)
    donut_frac = luigi.FloatParameter(default=0.2)
    log_donut = luigi.BoolParameter(default=False)
    max_with_lower_left = luigi.BoolParameter(default=False)
    log_transform = luigi.Parameter(default='auto')
    global_expected = luigi.BoolParameter(default=False)
    plot_outfile = luigi.Parameter(default=None)
    plot_outfile_kde = luigi.BoolParameter(default=False)
    plot_outfile_hexbin = luigi.BoolParameter(default=True)

    def _construct_cmd_string(self):
        cmd = 'lib5c expected -p %s -d %i -f %f -w %i -x %i -m %f -t %s' % \
              (self.input()[0].path, self.degree, self.lowess_frac, self.w,
               self.p, self.donut_frac, self.log_transform)
        if self.global_expected:
            cmd += ' -G'
        if self.monotonic:
            cmd += ' -M'
        if self.exclude_near_diagonal:
            cmd += ' -E'
        if self.powerlaw:
            cmd += ' -P'
        if self.regression:
            cmd += ' -R'
        if self.lowess:
            cmd += ' -L'
        if self.donut:
            cmd += ' -D'
        if self.log_donut:
            cmd += ' -O'
        if self.max_with_lower_left:
            cmd += ' -X'
        if self.plot_outfile is not None:
            if hasattr(self, 'rep'):
                cmd += ' -o %s' % self.plot_outfile.replace(r'%s', self.rep)
            else:
                cmd += ' -o %s' % self.plot_outfile
            if self.plot_outfile_kde:
                cmd += ' -K'
            if self.plot_outfile_hexbin:
                cmd += ' -B'
        cmd += ' %s %s' % (self.input()[1].path, self.output().path)
        return cmd


class VarianceTask(CmdTask):
    model = luigi.Parameter(default='lognorm')
    source = luigi.Parameter(default='local')
    min_obs = luigi.FloatParameter(default=2.0)
    min_disp = luigi.Parameter(default='1e-8')
    min_dist = luigi.IntParameter(default=6)
    x_unit = luigi.Parameter(default='dist')
    y_unit = luigi.Parameter(default='disp')
    logx = luigi.BoolParameter(default=False)
    logy = luigi.BoolParameter(default=False)
    regional = luigi.BoolParameter(default=False)

    def _construct_cmd_string_from_inpaths(self, obs_inpath, exp_inpath):
        cmd = 'lib5c variance -p %s -m %s -s %s --min_obs %s --min_disp %s ' \
              '--min_dist %s --x_unit %s --y_unit %s' % \
              (self.input()[0].path, self.model, self.source, self.min_obs,
               self.min_disp, self.min_dist, self.x_unit, self.y_unit)
        if self.logx:
            cmd += ' --logx'
        if self.logy:
            cmd += ' --logy'
        if self.regional:
            cmd += ' -R'
        cmd += ' %s %s %s' % (obs_inpath, exp_inpath, self.output().path)
        return cmd

    def _construct_cmd_string(self):
        return self._construct_cmd_string_from_inpaths(
            self.input()[1].path, self.input()[2].path)


class CrossVarianceTask(VarianceTask):
    source = luigi.Parameter(default='cross')
    conditions = luigi.Parameter(default='ES,pNPC')

    def _match_inputs(self):
        input_paths = [i.path for i in self.input()[2:]]
        common_prefix = os.path.commonprefix(input_paths)
        common_postfix = os.path.commonprefix(
            [p[::-1] for p in input_paths])[::-1]
        return common_prefix + '*' + common_postfix

    def _construct_cmd_string(self):
        return self._construct_cmd_string_from_inpaths(
            shell_quote(self._match_inputs()), self.input()[1].path)


@visualizable(pvalue=True)
class PvalueTask(CmdTask):
    distribution = luigi.Parameter(default='nbinom')
    log = luigi.BoolParameter(default=False)
    vst = luigi.BoolParameter(default=False)

    def _construct_cmd_string(self):
        cmd = 'lib5c pvalues -p %s %s %s %s %s %s' % \
              (self.input()[0].path, self.input()[1].path, self.input()[2].path,
               self.input()[3].path, self.distribution, self.output().path)
        if self.log:
            cmd += ' -L'
        if self.vst:
            cmd += ' -V'
        return cmd


@visualizable(tetris=True)
class ThresholdTask(CmdTask):
    kappa_confusion_outfile = luigi.Parameter(default='kappa_confusion.txt')
    dataset_outfile = luigi.Parameter(default='dataset.tsv')
    significance_threshold = luigi.FloatParameter(default=1e-15)
    bh_fdr = luigi.BoolParameter(default=False)
    concordant = luigi.BoolParameter(default=False)
    distance_threshold = luigi.IntParameter(default=24000)
    size_threshold = luigi.IntParameter(3)
    background_threshold = luigi.FloatParameter(default=0.6)
    conditions = luigi.Parameter(default='ES,pNPC')

    def _construct_cmd_string(self):
        return 'lib5c threshold ' \
            '-p %s -c %s -t %s %s %s -s %s -d %s -b %s -o %s -k %s %s %s' \
            % (self.input()[0].path, shell_quote(self.conditions),
               self.significance_threshold, '-B' if self.bh_fdr else '',
               '-C' if self.concordant else '', self.size_threshold,
               self.distance_threshold, self.background_threshold,
               self.output()[1].path, self.output()[2].path,
               self.output()[0].path,
               ' '.join([x.path for x in self.input()[1:]]))


@visualizable(pvalue=True)
class QvaluesTask(CmdTask):
    method = luigi.Parameter(default='bh_fdr')

    def _construct_cmd_string(self):
        cmd = 'lib5c qvalues -p %s -m %s %s %s' %\
              (self.input()[0].path, self.method, self.input()[1].path,
               self.output().path)
        return cmd


@visualizable(obs_over_exp=True)
class DivideTask(CmdTask):
    """
    Task class for dividing one countsfile by another.

    Input/output specification:
      * ``self.input()[0]``: the primer or bin .bed file
      * ``self.input()[1]``: the dividend (countsfile to divide)
      * ``self.input()[2]``: the divisor (countsfile to divide by)
      * ``self.output()``: the quotient (countsfile resulting from the division)
    """
    def _construct_cmd_string(self):
        cmd = 'lib5c divide -p %s %s %s %s' %\
              (self.input()[0].path, self.input()[1].path, self.input()[2].path,
               self.output().path)
        return cmd


class LogTask(CmdTask):
    """
    Task class for logging or unlogging a countsfile.

    Input/output specification:
      * ``self.input()[0]``: the primer or bin .bed file
      * ``self.input()[1]``: the input countsfile (to be logged)
      * ``self.output()``: the resulting countsfile (after logging)

    Attributes
    ----------
    log_base : str
        Base to log with.
    pseudocount : float
        Pseudocount to add before logging.
    unlog : bool
        Set to True to exponentiate instead of logging.
    """
    log_base = luigi.Parameter(default='e')
    pseudocount = luigi.FloatParameter(default=1)
    unlog = luigi.BoolParameter(default=False)

    def _construct_cmd_string(self):
        cmd = 'lib5c log -p %s %s -b %s -s %s %s %s' %\
              (self.input()[0].path, '-U' if self.unlog else '', self.log_base,
               self.pseudocount, self.input()[1].path, self.output().path)
        return cmd


@visualizable(obs_over_exp=True)
class InteractionScoreTask(CmdTask):
    """
    Task class for converting p-values to interaction scores.

    Input/output specification:
      * ``self.input()[0]``: the primer or bin .bed file
      * ``self.input()[1]``: the input countsfile of p-values
      * ``self.output()``: the resulting countsfile of interaction scores
    """
    def _construct_cmd_string(self):
        cmd = 'lib5c interaction-score -p %s %s %s' %\
              (self.input()[0].path, self.input()[1].path, self.output().path)
        return cmd


class DistributionTask(CmdTask):
    mode = luigi.Parameter(default='obs_over_exp')
    dist = luigi.Parameter(default='nbinom')
    log = luigi.BoolParameter(default=False)

    def _construct_cmd_string(self):
        raise NotImplementedError()


@visualizable(pvalue=True)
class LegacyPvaluesOneTask(DistributionTask):
    bias = luigi.BoolParameter(default=False)

    def _construct_cmd_string(self):
        cmd = 'lib5c pvalues -p %s -m %s -d %s' % \
              (self.input()[0].path, self.mode, self.dist)
        if self.log:
            cmd += ' -L'
        if self.bias:
            cmd += ' -b %s' % self.input()[1].path.replace('.counts', '.bias')
        cmd += ' %s %s %s' % \
               (self.input()[1].path, self.input()[2].path, self.output().path)
        return cmd


@visualizable(pvalue=True)
class LegacyPvaluesTwoTask(CmdTask):
    grouping = luigi.Parameter(default='distance')
    distance_tolerance = luigi.IntParameter(default=2)
    fractional_tolerance = luigi.FloatParameter(default=0.1)
    mode = luigi.Parameter(default='obs_over_exp')
    dist = luigi.Parameter(default='nbinom')
    log = luigi.BoolParameter(default=False)
    bias = luigi.BoolParameter(default=False)

    def _construct_cmd_string(self):
        cmd = 'lib5c pvalues2 -p %s -g %s -f %s -t %s -m %s -d %s' % \
              (self.input()[0].path, self.grouping, self.fractional_tolerance,
               self.distance_tolerance, self.mode, self.dist)
        if self.log:
            cmd += ' -L'
        if self.bias:
            cmd += ' -b %s' % self.input()[1].path.replace('.counts', '.bias')
        cmd += ' %s %s %s' % \
               (self.input()[1].path, self.input()[2].path, self.output().path)
        return cmd


class LegacyVisualizeFitTask(DistributionTask, RegionalTaskMixin):
    distance_scale = luigi.IntParameter(default=None)
    expected_value = luigi.FloatParameter(default=None)
    tolerance = luigi.FloatParameter(default=0.5)

    def _outfile_pattern(self):
        raise NotImplementedError()

    def _parameterized_outfile_pattern(self):
        # baseline outfile from pattern
        outfile = self._outfile_pattern()

        # replace parameter values
        if self.distance_scale is not None:
            outfile = outfile.replace('%d', '%i' % self.distance_scale)
        if self.expected_value is not None:
            outfile = outfile.replace('%e', '%g' % self.expected_value)
        if self.tolerance is not None:
            outfile = outfile.replace('%t', '%g' % self.tolerance)

        return outfile

    def _construct_cmd_string(self):
        cmd = 'lib5c visualize-fits -p %s -m %s -d %s' % \
              (self.input()[0].path, self.mode, self.dist)
        if self.log:
            cmd += ' -L'
        if self.distance_scale is not None:
            cmd += ' -s %i' % self.distance_scale
        if self.expected_value is not None:
            cmd += ' -e %g' % self.expected_value
        if self.tolerance is not None:
            cmd += ' -t %g' % self.tolerance
        if self.region is not None:
            cmd += ' -r %s' % self.region
        cmd += ' %s %s %s' % (self.input()[1].path,
                              self.input()[2].path,
                              self._parameterized_outfile_pattern())
        return cmd


class LegacyVisualizeVarianceTask(DistributionTask, RegionalTaskMixin):
    def _outfile_pattern(self):
        raise NotImplementedError()

    def _construct_cmd_string(self):
        cmd = 'lib5c visualize-variance -p %s -m %s -d %s' % \
              (self.input()[0].path, self.mode, self.dist)
        if self.log:
            cmd += ' -L'
        if self.region is not None:
            cmd += ' -r %s' % self.region
        cmd += ' %s %s %s' % (self.input()[1].path,
                              self.input()[2].path,
                              self._parameterized_outfile_pattern())
        return cmd
