import matplotlib as mpl

from lib5c.plotters.extendable import BaseExtendableHeatmap
from lib5c.util.bed import check_intersect


class BedExtendableHeatmap(BaseExtendableHeatmap):
    def add_bed_tracks(self, bed_tracks, size='3%', pad=0.0, axis_limits=(0, 1),
                       intron_height=0.05, exon_height=0.5, name=None,
                       colors=None, track_label=None):
        """ Runs add_bed_track using bed_tracks for both loc='bottom' and 'left'
        """
        ax_h = self.add_bed_track(
            bed_tracks, loc='bottom', size=size, pad=pad, name=name,
            axis_limits=axis_limits, intron_height=intron_height,
            exon_height=exon_height, colors=colors, track_label=track_label
        )
        ax_v = self.add_bed_track(
            bed_tracks, loc='left', size=size, pad=pad, name=name,
            axis_limits=axis_limits, intron_height=intron_height,
            exon_height=exon_height, colors=colors, track_label=track_label
        )
        return [ax_h, ax_v]

    def add_bed_track(self, bed_tracks, loc='bottom', size='3%', pad=0.0,
                      name='bed', axis_limits=(0, 1), intron_height=0.05,
                      colors=None, track_label=None):
        """ Adds iterable corresponding to BED file format as track to plot

        Parameters
        ----------
        bed_tracks : Iterable[?]
        loc : Optional[str]
            Where the track is being added, goes into self.add_margin_ax and
            used to rotate axes as necessary
        size : TODO
        pad : TODO
        name : TODO
        axis_limits : TODO
        intron_height : TODO
        colors : Dict[str, str]
            Map from strand value to color name for bed_tracks with that strand
            value (e.g. orientation). If not provided for a given strand,
            color is 'k' by default.
        track_label : Optional[str]
            If specified, label the track with this value
        """
        # deduce orientation, either h or v, and save the correct grange
        if loc in ['bottom', 'top']:
            orientation = 'horizontal'
            grange = self.grange_x
        else:
            orientation = 'vertical'
            grange = self.grange_y

        # create new axis
        ax = self.add_margin_ax(loc=loc, size=size, pad=pad,
                                new_ax_name="{}_{}".format(orientation, name),
                                axis_limits=axis_limits)

        # deduce midpoint of the non-genomic axis
        midpoint = (axis_limits[0] + axis_limits[1]) / 2.

        # compute track_width and track_height, assuming horizontal orientation
        track_width = grange['end'] - grange['start']

        # create patches for each bed
        for bed in bed_tracks:
            # skip bed_tracks that don't intersect the window
            if not check_intersect(bed, grange):
                continue

            # get the color for this bed
            try:
                color = colors[bed['strand']]
            except (TypeError, KeyError):
                color = 'k'

            # plot small rectangle for length of bed
            intron_coords = [bed['start'], midpoint - intron_height / 2.]
            intron_dims = [bed['end'] - bed['start'], intron_height]
            if orientation == 'vertical':
                intron_coords.reverse()
                intron_dims.reverse()
            ax.add_artist(mpl.patches.Rectangle(intron_coords, *intron_dims,
                                                fc=color, ec=color))
        # done looping over all bed_tracks

        # write label of bed track, if specified
        if track_label:
            name_coords = [grange['start'] + track_width / 80, axis_limits[1]]
            ha = 'left'
            va = 'top'
            rotation = 0
            if orientation == 'vertical':
                name_coords.reverse()
                ha = 'right'
                rotation = 270
            ax.text(name_coords[0], name_coords[1], track_label, fontsize=7,
                    ha=ha, va=va, rotation=rotation)

        # return axis for this track
        return ax
